//------------------------------------------------------------------------
//
// OpenDCC - DCC Sniffer
//
// Copyright (c) 2008 Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
// 
//------------------------------------------------------------------------
//
// file:      parser.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// webpage:   http://www.opendcc.de
// history:   2008-03-02 V0.01 kw started
//            2008-04-18 V2.00 kw Tested with HTerm
//            2008-05-01 V2.01 kw added drop to lf if cr-lf is received;
//                                accept both cr, lf and cr-lf
//            2009-06-26 V2.02 kw added fask clock decoding
//
//------------------------------------------------------------------------
//
// purpose:   DCC sniffer
//            here: scan commmands on the serial interface and set 
//            internal states accordingly
//
// content:   1. Interface
//               calls for on, off, blinking, flashing, ...
//            2  runtime-part
//               must be called every time_tick
//
//------------------------------------------------------------------------

#include <stdlib.h>
#include <stdbool.h>
#include <inttypes.h>
#include <ctype.h>
#include <avr/pgmspace.h>        // put var to program memory
#include <avr/io.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <string.h>

#include "config.h"
#include "hardware.h"
#include "rs232.h"
#include "led_control.h"
#include "rprintf.h"
#include "tracker.h"             // get data definitions from tracker

#define SIMULATION  0            // 0: real application
                                 // 1: test receive routine
                                 // 2: test timing engine
                                 // 3: test action

//=============================================================================
// 1. Definitions
//=============================================================================
//
//
// #define LED_PROG     0     // program   - yellow
// #define LED_RAILCOM  1     // railcom   - yellow
// #define LED_ACC      2     // accessory - green
// #define LED_LOCO     3     // loco      - green
// #define LED_GO       4     // sniffing  - blue
// #define LED_ON       5     // on        - green




//----------------------------------------------------------------------------
//
// Variables and Tools for the Parser
//
// Im Gegensatz zu sonstigen C Gepflogenheiten halten wir den Pointer auf pcc
// global; damit kann der String mit einem Pointer durchsucht werden.
// 
unsigned char pcc[64];          // pc_message speicher

unsigned char pcc_size;         // if known: size of current command (for p50x_binary)
unsigned char pcc_index;        // current write or read index

unsigned char *ibcmd;           // globaler Pointer auf commandstring

// each command consists of a string (or char), follow by +, -, numeric or none
enum 
  {
    NONE,       // no parameters given
    NUMERIC,    // numeric parameter, value in CmdValue
    ENABLE,     // +
    DISABLE     // -
  }
  CmdQualifier;

unsigned int CmdValue;

enum
  {
    IDLE,
    WAIT_FOR_COMMAND,
    WAIT_FOR_PARAM
  } parser_state;



//------------------------------------------------------------------------
// 2. Command Parser
//------------------------------------------------------------------------
//
// a) run_parser: collects a complete input command and starts to parse it,
// when a terminating white char is detected.
//
// b) parse: checks for opcodes and parameter
//  and calls the corresponding function
//
// c) PA_xx: this are the functions
//

//--- support routine


unsigned char overwrite;

void add_tracker(t_trackpara *track, unsigned int new_tracker)
  { 
    unsigned char i;

    for (i=0; i<4; i++)
      {
        if (new_tracker == track->addr[i] )
        {
            track->bits |= (1<<i);
            return;   // tracker already existing, now activated
        }
      }

    for (i=0; i<4; i++)
      {
        if ( (track->bits & (1<<i)) == 0)
          {
            track->addr[i] = new_tracker;
            track->bits |= (1<<i);
            return;   //new tracker created
          }
      }

    overwrite = (overwrite + 1) % 4;
    i = overwrite;

    track->addr[i] = new_tracker;
    track->bits |= (1<<i);
    return;   //replaced oldest tracker
  }


unsigned char should_we_track_this(t_trackpara *track, unsigned int addr)
  {
    if ( track->bits  & 0x80) return(TRUE);
    if ((track->bits  & (1<<0)) && (track->addr[0] == addr)) return(TRUE);
    if ((track->bits  & (1<<1)) && (track->addr[1] == addr)) return(TRUE);
    if ((track->bits  & (1<<2)) && (track->addr[2] == addr)) return(TRUE);
    if ((track->bits  & (1<<3)) && (track->addr[3] == addr)) return(TRUE);
    return(FALSE);
  }


//------------------------------------------------------------------------

void PA_configure(void)
  {
    // right now, we don't reconfigure 
  }

void PA_verbose(void)
  {
    if (CmdQualifier == DISABLE)
      {
        SnifferState.DisplayMode &= ~D_VERBOSE;   // reset
      }
    else if (CmdQualifier == ENABLE)
      {
        SnifferState.DisplayMode |= D_VERBOSE;    // set 
      }
    else if (CmdQualifier == NUMERIC)
      {
        if (CmdValue == 1)  SnifferState.DisplayMode |= D_VUMETER;
        else SnifferState.DisplayMode &= ~D_VUMETER; 
      }
    else 
      {
        // ignore if any other parameter (should we do a toggle?)
        if (SnifferState.DisplayMode & D_VERBOSE) SnifferState.DisplayMode &= ~D_VERBOSE;   // reset
        else SnifferState.DisplayMode |= D_VERBOSE;    // set 
      }
  }



char Version[]  PROGMEM = "OpenDCC Xpressnet Sniffer - Version: 1.02, 31.10.2010";

char HL01[] PROGMEM = "Commands: <letter>[num][+|-] <ENTER>";
char HL02[] PROGMEM = "V[+|-] Display messages in verbose mode (able to be interpreted by humans)";
char HL03[] PROGMEM = "H[+|-] Display messages as hex bytes";
char HL04[] PROGMEM = "H0     Hex with no spaces, no xor             (:0368)";
char HL05[] PROGMEM = "H1     Hex with spaces between bytes, no xor  (:03 68)";
char HL06[] PROGMEM = "H2     Hex with no spaces, with xor           (:03686B)";
char HL07[] PROGMEM = "H3     Hex with spaces between bytes, with xor(:03 68 6B)";
char HL08[] PROGMEM = "A[+|-] Display acknowledge messages";
char HL09[] PROGMEM = "I<tp>  Display inquiry messages";
char HL10[] PROGMEM = "M<tp>  Display normal messages";
char HL11[] PROGMEM = "B[+|-] Display BiDi / feedback";
char HL12[] PROGMEM = "F[+|-] Display future / feedback";
char HL13[] PROGMEM = "T[+|-] Display timestamp; unit: 1ms ex: 123,567s";
char HL14[] PROGMEM = "TC[+|-]Display Fast Clock messages";
char HL15[] PROGMEM = "+      Display all messages";
char HL16[] PROGMEM = "-      Display no messages";
char HL17[] PROGMEM = "TRG    [+,-,R,F] Display external Triggerevents ";
char HL18[] PROGMEM = "?      Display list of commands supported";
char HL19[] PROGMEM = "??     Display current status";
char HL20[] PROGMEM = "<tp>   Tracker parameters (up to 4 trackers):";
char HL21[] PROGMEM = "        +=all; -=none; <num>: display only command with this address";
char HL22[] PROGMEM = "The sniffer remembers the last display mode at power up";


PGM_P HelpMsgArray[] PROGMEM = 
{
	HL01, HL02, HL03, HL04, HL05, HL06, HL07, HL08, HL09, HL10,  
	HL11, HL12, HL13, HL14, HL15, HL16, HL17, HL18, HL19, HL20, 
    HL21, HL22,  
} ;


void PA_Help(void)
  {
    rprintfProgStr( Version );
    rprintfCRLF() ;

    uint8_t i ;
    for( i = 0; i < ( sizeof( HelpMsgArray ) / sizeof( PGM_P ) ); i++ )
      {
        rprintfProgStr( (PGM_P)pgm_read_word( &(HelpMsgArray[i]) ) ) ;
        rprintfCRLF() ;
      }
  }

void PA_unknown(void)
  {
    toggle_EnableDisplay();
  }


void PA_hex(void)         
  {
    if (CmdQualifier == DISABLE)
      {
        SnifferState.DisplayMode &= ~D_HEX;   // reset
      }
    else if (CmdQualifier == ENABLE)
      {
        SnifferState.DisplayMode |= D_HEX;    // set 
      }
    else if (CmdQualifier == NUMERIC)
      {
        SnifferState.DisplayMode &= 0xf8;          // clear current HexMode
        SnifferState.DisplayMode |= CmdValue & 0x07;
        SnifferState.DisplayMode |= D_HEX;    // set 
      }
    else
      {
        if (SnifferState.DisplayMode & D_HEX) SnifferState.DisplayMode &= ~D_HEX;   // reset
        else SnifferState.DisplayMode |= D_HEX;    // set 
      }
  }



void PA_inquiry(void) 
  {
    if (CmdQualifier == NUMERIC)
      {
         add_tracker (&SnifferState.TrackInquiry, CmdValue);
      }
    else if (CmdQualifier == DISABLE)
      {
        SnifferState.TrackInquiry.bits = 0;   // reset all
      }
    else if (CmdQualifier == ENABLE)
      {
        SnifferState.TrackInquiry.bits = (1<<7);   // set MSB
      }
    else 
      {
        SnifferState.TrackInquiry.bits = (1<<7);   // default - same as ON
      }
  }

void PA_message(void) 
  {
    if (CmdQualifier == NUMERIC)
      {
         add_tracker (&SnifferState.TrackMessage, CmdValue);
      }
    else if (CmdQualifier == DISABLE)
      {
        SnifferState.TrackMessage.bits = 0;   // reset all
      }
    else if (CmdQualifier == ENABLE)
      {
        SnifferState.TrackMessage.bits = (1<<7);   // set MSB
      }
    else 
      {
        SnifferState.TrackMessage.bits = (1<<7);   // default - same as ON
      }
  }


 
void PA_acknowledge(void)
  {
    if (CmdQualifier == DISABLE)
      {
        SnifferState.TrackAck = 0;   // reset
      }
    else if (CmdQualifier == ENABLE)
      {
        SnifferState.TrackAck = (1<<7);   // set MSB
      }
    else 
      {
        // ignore if any other parameter (should we do a toggle?)
      }
  }

static void PA_BiDi(void)
  {
    if (CmdQualifier == DISABLE)
      {
        SnifferState.TrackBiDi = 0;   // reset
      }
    else if (CmdQualifier == ENABLE)
      {
        SnifferState.TrackBiDi = (1<<7);   // set MSB
      }
    else 
      {
        // ignore if any other parameter (should we do a toggle?)
      }
  }
         


void PA_timestamp(void)
  {
    if (CmdQualifier == DISABLE)
      {
        SnifferState.DisplayMode &= ~D_TSTAMP;   // reset
      }
    else if (CmdQualifier == ENABLE)
      {
        SnifferState.DisplayMode |= D_TSTAMP;    // set 
      }
    else 
      {
        // ignore if any other parameter (should we do a toggle?)
        if (SnifferState.DisplayMode & D_TSTAMP) SnifferState.DisplayMode &= ~D_TSTAMP;   // reset
        else SnifferState.DisplayMode |= D_TSTAMP;    // set 

      }
  }


static void PA_fastclock(void)
  {
    if (CmdQualifier == DISABLE)
      {
        SnifferState.TrackFastClk = 0;   // reset
      }
    else if (CmdQualifier == ENABLE)
      {
        SnifferState.TrackFastClk = (1<<7);   // set MSB
      }
    else 
      {
        // ignore if any other parameter (should we do a toggle?)
      }
  }


static void PA_trig_rising(void)
  {
    SnifferState.TrackTrigger = D_ACTIVE; 
  }

static void PA_trig_falling(void)
  {
    SnifferState.TrackTrigger = D_INACTIVE; 
  }

static void PA_trig(void)
  {
    if (CmdQualifier == DISABLE)
      {
        SnifferState.TrackTrigger = 0;   // reset
      }
    else if (CmdQualifier == ENABLE)
      {
        SnifferState.TrackTrigger = D_ACTIVE | D_INACTIVE; 
      }
    else 
      {
        // ignore if any other parameter (should we do a toggle?)
      }
  }



static void PA_full(void)
  {
    SnifferState.TrackAck = (1<<7); 
    SnifferState.TrackFuture = (1<<7); 
    SnifferState.TrackInquiry.bits = (1<<7); 
    SnifferState.TrackMessage.bits = (1<<7); 
  }

static void PA_none(void)
  {
    SnifferState.TrackAck = 0; 
    SnifferState.TrackFuture = 0; 
    SnifferState.TrackInquiry.bits = 0; 
    SnifferState.TrackMessage.bits = 0; 
  }


void showstat(unsigned char state)
  {
    if (state) rprintfChar( '+' ) ;
    else rprintfChar( '-' ) ;
  }

void showtracker(t_trackpara *tracker)
  {
    unsigned char i;
    if (tracker->bits & 0x80) rprintfChar( '+' ) ;
    else if (tracker->bits == 0) rprintfChar( '-' ) ;
    else
        for (i=0; i<4; i++)
          {
            if (tracker->bits & (1<<i))
              {
                rprintfNum( 10, 5, 0, '0', tracker->addr[i] ) ;
                rprintfChar(' ');
              }
          }
  }


char text_state[]      PROGMEM = "State:     "  ;
char text_display[]    PROGMEM = "Display:   "  ;
char text_inquiry[]    PROGMEM = "Inquiry:   "  ;
char text_message[]    PROGMEM = "Message:   "  ;
char text_acks[]       PROGMEM = "Acks:      "  ;
char text_bidi[]       PROGMEM = "BiDi:      "  ;
char text_future[]     PROGMEM = "Future:    "  ;
char text_trig[]       PROGMEM = "Ext.Trig:  "  ;
char text_on[]         PROGMEM = "on"  ;
char text_off[]        PROGMEM = "off"  ;


void PA_showstate(void)
  {
    rprintfProgStr( Version );
    rprintfCRLF() ;
    rprintfProgStr( text_state ) ;
    if (EnableDisplay) rprintfProgStr( text_on );
    else  rprintfProgStr( text_off );
    rprintfCRLF() ;

    rprintfProgStr( text_display ) ;
    rprintfChar( 'T' ) ;
    showstat( (SnifferState.DisplayMode & D_TSTAMP)  ) ;
    rprintfChar( 'H' ) ;
    showstat( (SnifferState.DisplayMode & D_HEX)  ) ;
    rprintfChar( 'V' ) ;
    showstat( (SnifferState.DisplayMode & D_VERBOSE)  ) ;
    rprintfCRLF() ;

    rprintfProgStr( text_inquiry ) ;
    showtracker(&SnifferState.TrackInquiry);
    rprintfCRLF() ;

    rprintfProgStr( text_message ) ;
    showtracker(&SnifferState.TrackMessage);
    rprintfCRLF() ;

    rprintfProgStr( text_acks ) ;
    showstat(SnifferState.TrackAck) ;
    rprintfCRLF() ;

    rprintfProgStr( text_bidi ) ;
    showstat(SnifferState.TrackBiDi) ;
    rprintfCRLF() ;

    rprintfProgStr( text_future ) ;
    showstat(SnifferState.TrackFuture) ;
    rprintfCRLF() ;

    rprintfProgStr( text_trig ) ;
    if (SnifferState.TrackTrigger == 0) rprintfChar('-');
    if (SnifferState.TrackTrigger & D_ACTIVE) rprintfChar('R');
    if (SnifferState.TrackTrigger & D_INACTIVE) rprintfChar('F');
    rprintfCRLF() ;
  }


// debug

void PA_xxx(void)
  {
    if (CmdQualifier == NUMERIC)
      {
      }
    else if (CmdQualifier == DISABLE)
      {
        X_push_to_rx(0x100 | 0xA1);      // feedback
        X_push_to_rx(0x42);
        X_push_to_rx(0x03);
        X_push_to_rx(0x05);
        X_push_to_rx(0x44);      }
    else if (CmdQualifier == ENABLE)
      {
        X_push_to_rx(0x100 | 0xA0);      // feedback
        X_push_to_rx(0x42);
        X_push_to_rx(0x03);
        X_push_to_rx(0x05);
        X_push_to_rx(0x44);
      }
    else 
      {
      }
  }

//------------------------------------------------------------------------------
// Hilfsrountinen:

signed char char2hex (unsigned char *arg)
  {
    if ((*arg >= '0') && (*arg <= '9')) return (*arg - '0');
    if ((*arg >= 'A') && (*arg <= 'F')) return (*arg - 'A' + 10);
    if ((*arg >= 'a') && (*arg <= 'f')) return (*arg - 'a' + 10);
    return (-1);  // no hex
  }

signed char char2bin (unsigned char *arg)
  {
    if (*arg == '0') return(0);
    if (*arg == '1') return(1);
    return (-1);  // no bin
  }

/*
// Standard C Konventions (not used in OPENDCC)
// axtoi bricht beim ersten non-hex-Wert ab

unsigned int axtoi(char *hexStg)
  {
    unsigned char n = 0;            // position in string
    unsigned int intValue = 0;      // integer value of hex string
    signed char temp;               // convertresult
    

    while (n < 16)                 // security limit against wrong string    
      {
        if ((temp=char2hex(hexStg[n++])) < 0) return(intValue);
        intValue = (intValue << 4) + temp;
      }
    return (intValue);
  }

unsigned int abtoi(char *binStg)
  {
    unsigned char n = 0;            // position in string
    unsigned int intValue = 0;      // integer value of bin string
    signed char temp;               // convertresult
    

    while (n < 32)                 // security limit against wrong string    
      {
        if ((temp=char2bin(binStg[n++])) < 0) return(intValue);
        intValue = (intValue << 1) + temp;
      }
    return (intValue);
  }
*/
//-------------------------------------------------------------
// Diese Funktionen nochmal, jetzt als Spezialversion mit
// globalen Variablen; Der pcc_index bleibt auf dem ersten falschen
// Zeichen stehen.

// axtoi bricht beim ersten non-hex-Wert ab

unsigned int get_cmdvalx(void)
  {
    unsigned int intValue = 0;      // integer value of hex string
    signed char temp;               // convertresult
    

    while (*ibcmd != 0)                 // security limit against wrong string    
      {
        if ((temp=char2hex(ibcmd)) < 0) return(intValue);
        intValue = (intValue << 4) + temp;
        ibcmd++;
      }
    return (intValue);
  }

unsigned int get_cmdvalb(void)
  {
    unsigned int intValue = 0;      // integer value of bin string
    signed char temp;               // convertresult
    

    while (*ibcmd != 0)             
      {
        if ((temp=char2bin(ibcmd)) < 0) return(intValue);
        intValue = (intValue << 1) + temp;
        ibcmd++;
      }
    return (intValue);
  }
            

void parse_parameter(void)
  {
    CmdQualifier = NONE;
    while (*ibcmd == ' ') ibcmd++;      // skip blanks

    if (*ibcmd == 0)
      {
        return;  // empty arg, end of line!
      }
    if (*ibcmd == '+')
      {
        CmdQualifier = ENABLE;
        return;  
      }
    if (*ibcmd == '-')
      {
        CmdQualifier = DISABLE;
        return;  
      }
    if ((*ibcmd >= '0') && (*ibcmd <= '9'))
      {
        CmdQualifier = NUMERIC;
        CmdValue = atoi((char *)ibcmd);
        // while ((*ibcmd >= '0') && 
        //       (*ibcmd <= '9'))  ibcmd++;  // skip numbers
        // if (*ibcmd == ',') ibcmd++;   // skip ',' 
      }
    if (*ibcmd == '$')
      {
        CmdQualifier = NUMERIC;
        ibcmd++;
        CmdValue = get_cmdvalx();
      }    
  }


// Note: the scan stops after the first match, therefore a single char command
//       must be the last command with this character.
//       There must be an empty command at the end of the list - otherwise the
//       scan will not terminate.

typedef struct
{
  unsigned char command[5];
  void (*fnAction)(void);
} T_AACTION;

T_AACTION AsciiParseTab[] PROGMEM =
  {
    // Command / Call             // implemented / simulated / tested / comment 
    {{"C"},    PA_configure },    // 
    {{"V"},    PA_verbose },      // i /
    {{"HELP"}, PA_Help },         // i / s / t / help string
    {{"H"},    PA_hex },          // i / -         
    {{"A"},    PA_acknowledge },  // i / -         
    {{"B"},    PA_BiDi },         // i / -         
    {{"I"},    PA_inquiry },      // i / -         
    {{"M"},    PA_message},       // i / -
    {{"+"},    PA_full },         // n / - 
    {{"-"},    PA_none},          // i / - / - /
    {{"TC"},   PA_fastclock },    // i / -         
    {{"TRGR"}, PA_trig_rising},   // i / - /
    {{"TRGF"}, PA_trig_falling},  // i / - /
    {{"TRG"},  PA_trig},          // i / - / 
    {{"T"},    PA_timestamp},     // i / - / t / +/- a timestamp [1ms]
    {{"X"},    PA_xxx},    
    {{"??"},   PA_showstate },    // i / s / - /
    {{"?"},    PA_Help },         // i / s / - /
    {{""},     PA_unknown},       // i / s / - / empty command is always true
};

//--------------------------------------------------------------------------------
// This routine checks for opcodes and parameter
//  and calls the corresponding function
//
void parse (void)
  {
    unsigned char *arg;
    unsigned char i, j;
    void (*fp)(void);                   // declare pointer to function


    j=0;
    while(1)
      {
        arg = ibcmd;                    // reset pointer to arg
        for (i=0; i<5; i++)
          {
            if (pgm_read_byte(&AsciiParseTab[j].command[i]) == 0)   // target found
              {
                ibcmd = ibcmd+i;        // advance parse pointer to arguments

                parse_parameter();
                
                fp = (void *) pgm_read_word(&AsciiParseTab[j].fnAction);   // assign function addr read from flash
                (*fp)();                  // now call this function
                 
                save_state_changes();        
                return;
              }
            if (*arg++ != pgm_read_byte(&AsciiParseTab[j].command[i]) )
              {
                break;                // this target is wrong
              }
          }
        j++;
      }
  }
 

//------------------------------------------------------------------------
// 

void init_parser(void)
  {
    parser_state = IDLE;
  }


bool input_ready(void)
  {
    if(rs232_break_detected == 1)
      {
        init_parser();
        init_rs232(DEFAULT_BAUD);  // reinit rs232 and get rid of old connection
      }
    return(rx_fifo_ready());
  }


// new: drop lf after cr
unsigned char last_received;

unsigned char drop_linefeed_after_cr(unsigned char test)
  {
    unsigned char retval = FALSE;
    if (test == 0x0a)
      {
        if (last_received == 0x0d)          // if <cr><lf> -> drop <lf>
          {
            retval = TRUE;
          }
      }
    last_received = test;
    return(retval);
  }


unsigned char input_read(void)
  {
    return(rx_fifo_read());
  }



void run_parser(void)
  {
    unsigned char received;

    switch (parser_state)
      {
        case IDLE:                                      
            parser_state = WAIT_FOR_COMMAND;
            CmdQualifier = NONE;
            pcc_index = 0;
            ibcmd = &pcc[0];
            break;
        case WAIT_FOR_COMMAND:
            if (!input_ready()) return;
            
            received = input_read();
            if (drop_linefeed_after_cr(received))  return;
            pcc_index = 1;
            parser_state = WAIT_FOR_PARAM;
            *ibcmd = toupper(received);
            break;
        case WAIT_FOR_PARAM:
            if ((*ibcmd == 0x0a) || (*ibcmd == 0x0d))       // terminator
              {
                *ibcmd = 0;                                 // terminate string
                ibcmd = &pcc[0];
                parse();
                parser_state = IDLE;
                return;
              }
            if (!input_ready()) return;
            received = input_read();
            if (drop_linefeed_after_cr(received))  return;
                 
            if (pcc_index < (64-1))
              {
                pcc_index++;
                ibcmd++;                                    // advance read pointer
              }
            else 
              {
                // - fatal error - string too long (keep only last char)
              }
            *ibcmd = toupper(received);                    // read char and dump to command buffer
            break;
     }
  }






